import java.util.Vector; 
import java.io.*; 
import java.util.Collections; 

/* Copyright K Lano 2012-2019
   package: TransformationSynthesis
*/ 


public class AttributeMatching
{ Attribute src; 
  Attribute trg; 
  String srcname = ""; 
  String trgname = ""; 
  Expression srcvalue = null; // in cases where a value is mapped to the target
  EntityMatching dependsOn;  // in cases where src : E or Collection(E) and E maps to F
  Attribute elementVariable = null; // for  expr -> trg
  Vector auxVariables = new Vector();     // for  expr -> trg


  public AttributeMatching(Attribute source, Attribute target)
  { src = source; 
    trg = target; 
    srcname = src.getName(); 
    trgname = trg.getName(); 
  } 

  public AttributeMatching(Expression source, Attribute target)
  { src = new Attribute("" + source, source.getType(), 
                        ModelElement.INTERNAL); 
    srcvalue = source; 
    trg = target; 
    srcname = srcvalue + ""; 
    trgname = trg.getName(); 
  } 

  public AttributeMatching(Expression source, Attribute target, Attribute var, 
                           Vector auxvars) 
  { src = new Attribute("" + source, new Type("String",null), 
                        ModelElement.INTERNAL); 
    srcvalue = source; 
    trg = target; 
    elementVariable = var; 
    srcname = srcvalue + ""; 
    trgname = trg.getName();
    auxVariables = auxvars;  
  } 

  public String toString()
  { return "    " + srcname + " |--> " + trgname; } 

  public Vector inverts()
  { Vector res = new Vector(); 
    if (isStringAssignment())
    { // assume auxVariables.size() == 2
      BasicExpression delim = new BasicExpression("\"~\""); 
      delim.setType(new Type("String",null)); 
      BasicExpression etrg = new BasicExpression(trg); 
      BinaryExpression be1 = new BinaryExpression("->before", etrg, delim); 
      BinaryExpression be2 = new BinaryExpression("->after", etrg, delim); 
      Attribute v1 = (Attribute) auxVariables.get(0); 
      Attribute v2 = (Attribute) auxVariables.get(1); 
      res.add(new AttributeMatching(be1,v1)); 
      res.add(new AttributeMatching(be2,v2)); 
      return res; 
    } 
    else if (isExpressionAssignment())
    { if (srcvalue instanceof SetExpression) 
      { SetExpression sse = (SetExpression) srcvalue; 
        if (sse.isOrdered() && sse.size() == 2 && trg.isOrdered())
        { // Sequence{r1,r2} --> trg inverts to trg->at(1) --> r1, trg->at(2) --> r2
          Expression v1 = sse.getExpression(0); 
          Expression v2 = sse.getExpression(1); 
          Attribute a1 = new Attribute(v1 + "",v1.getType(),
                                       ModelElement.INTERNAL); 
          Attribute a2 = new Attribute(v2 + "",v2.getType(),
                                       ModelElement.INTERNAL);
          BasicExpression ind1 = new BasicExpression(1); 
          BasicExpression ind2 = new BasicExpression(2);
          BasicExpression trge = new BasicExpression(trg); 
          trge.setUmlKind(Expression.ATTRIBUTE); 

          BinaryExpression at1 = new BinaryExpression("->at",trge,ind1); 
          at1.setType(trg.getElementType()); 
          at1.setElementType(trg.getElementType()); 

          BinaryExpression at2 = new BinaryExpression("->at",trge,ind2); 
          at2.setType(trg.getElementType()); 
          at2.setElementType(trg.getElementType()); 

          AttributeMatching am1 = new AttributeMatching(at1,a1); 
          AttributeMatching am2 = new AttributeMatching(at2,a2);                     
          res.add(am1); 
          res.add(am2); 
          return res; 
        }
      }            

      // if r1->including(r2) --> trg with trg ordered then invert to 
      // trg->front() --> r1, trg->last() --> r2
      if (srcvalue instanceof BinaryExpression) 
      { BinaryExpression sbe = (BinaryExpression) srcvalue; 
        if (sbe.operator.equals("->including") && 
            sbe.left instanceof BasicExpression && 
            sbe.right instanceof BasicExpression && trg.isOrdered()) 
        { BasicExpression lft = (BasicExpression) sbe.left; 
          BasicExpression rgt = (BasicExpression) sbe.right; 
          Attribute a1 = new Attribute(lft.data,lft.getType(),
                                       ModelElement.INTERNAL); 
          Attribute a2 = new Attribute(rgt.data,rgt.getType(),
                                       ModelElement.INTERNAL); 
          Expression trge = new BasicExpression(trg); 
          UnaryExpression at1 = new UnaryExpression("->front",trge); 
          UnaryExpression at2 = new UnaryExpression("->last",trge); 
          res.add(new AttributeMatching(at1,a1)); 
          res.add(new AttributeMatching(at2,a2)); 
          return res; 
        }
      }  
      
      // if r1->union(r2) --> trg with trg or r1 being sets then invert to 
      // trg --> r1, trg --> r2
      if (srcvalue instanceof BinaryExpression) 
      { BinaryExpression sbe = (BinaryExpression) srcvalue; 
        if (sbe.operator.equals("->union") && 
            sbe.left instanceof BasicExpression && 
            sbe.right instanceof BasicExpression && 
            !(sbe.left.isOrdered() && trg.isOrdered())) 
        { BasicExpression lft = (BasicExpression) sbe.left; 
          BasicExpression rgt = (BasicExpression) sbe.right; 
          Attribute a1 = new Attribute(lft.data,lft.getType(),
                                       ModelElement.INTERNAL); 
          Attribute a2 = new Attribute(rgt.data,rgt.getType(),
                                       ModelElement.INTERNAL); 
          res.add(new AttributeMatching(trg,a1)); 
          res.add(new AttributeMatching(trg,a2)); 
          return res; 
        }
      }  
    } 

          
    AttributeMatching invam = invert(); 
    res.add(invam); 
    return res; 
  } 
      
  public AttributeMatching invert()
  { return new AttributeMatching(trg,src); } 
  // ignores any expression/value match

  boolean isDirect() // either source or target is a direct attribute
  { if (src.getNavigation().size() <= 1) 
    { return true; } 
    if (trg.getNavigation().size() <= 1) 
    { return true; } 
    return false; 
  } 

  boolean isDirectSource() // source is a direct attribute
  { if (src.getNavigation().size() <= 1) 
    { return true; } 
    return false; 
  } 

  boolean isDirectTarget() // target is a direct attribute
  { if (trg.getNavigation().size() <= 1) 
    { return true; } 
    return false; 
  } 

  boolean isValueAssignment() 
  { return srcvalue != null && elementVariable == null; } 

  boolean isExpressionAssignment()
  { return srcvalue != null && elementVariable != null; } 

  boolean isStringAssignment() 
  { return srcvalue != null && auxVariables.size() > 1; } 

  Expression addReference(BasicExpression ref, Type t)
  { if (srcvalue != null) 
    { return srcvalue.addReference(ref,t); } 
    return null; 
  } 

  public Vector getAuxVariables()
  { return auxVariables; } 

  public Expression getExpression()
  { return srcvalue; } 

  public Attribute getElementVariable()
  { return elementVariable; } 

  static Expression dataConversion(Expression sexp, Type stype, Type ttype)
  { // Convert sexp : stype to ttype
    if (("" + stype).equals("" + ttype)) 
    { return sexp; } 

    String sname = stype.getName(); 
    String tname = ttype.getName(); 
    Type selemt = stype.getElementType(); 
    Type telemt = ttype.getElementType(); 

    if (("" + selemt).equals("" + telemt))
    { if (sname.equals("Sequence") && tname.equals("Set"))
      { UnaryExpression res = new UnaryExpression("->asSet", sexp); 
        return res; 
      } 
      else if (sname.equals("Set") && tname.equals("Sequence"))
      { UnaryExpression res = new UnaryExpression("->asSequence", sexp); 
        return res; 
      }
      else if (sname.equals("Set") && !ttype.isCollectionType())
      { UnaryExpression res = new UnaryExpression("->any", sexp); 
        return res; 
      }  
      else if (sname.equals("Sequence") && !ttype.isCollectionType())
      { UnaryExpression res = new UnaryExpression("->any", sexp); 
        return res; 
      }  
      else if (tname.equals("Set") && !stype.isCollectionType())
      { SetExpression res = new SetExpression();
        res.addElement(sexp);  
        return res; 
      }  
      else if (tname.equals("Sequence") && !stype.isCollectionType())
      { SetExpression res = new SetExpression();
        res.setOrdered(true); 
        res.addElement(sexp);  
        return res; 
      }  
    } 
    
    if (stype.isEnumeration() && tname.equals("String"))
    { String query = sname + "2String";  
      BasicExpression res = new BasicExpression(query); 
      res.setIsEvent(); 
      res.setUmlKind(Expression.QUERY); 
      res.addParameter(sexp); 
      return res; 
    } 
    else if (ttype.isEnumeration() && sname.equals("String"))
    { String query = "String2" + tname;  
      BasicExpression res = new BasicExpression(query); 
      res.setIsEvent(); 
      res.setUmlKind(Expression.QUERY); 
      res.addParameter(sexp); 
      return res; 
    } 
    

    if (stype.isCollectionType() && !ttype.isCollectionType())
    { UnaryExpression res = new UnaryExpression("->any", sexp); 
      return res; 
    } 
    else if (ttype.isCollectionType() && !stype.isCollectionType())
    { UnaryExpression res = new UnaryExpression("->as" + sname, sexp); 
      return res; 
    } 
 
    return new BinaryExpression("->oclAsType", sexp, new BasicExpression(ttype)); 
  } 

  public Vector analyseCorrelationPatterns(Entity s, Entity t, Entity realsrc, Entity realtrg, 
                                           Vector ems) 
  { Vector res = new Vector(); 

    if (src.isComposition() && !trg.isComposition()) 
    { CorrelationPattern p = new CorrelationPattern("Remove aggregation", 
                                                    "Aggregation " + realsrc + "::" + src + 
                                                    " maps to non-aggregation " + 
                                                    realtrg + "::" + trg); 
      p.addSourceEntity(realsrc); 
      p.addTargetEntity(realtrg); 
      p.addSourceFeature(src); 
      p.addTargetFeature(trg); 
      if (res.contains(p)) { } 
      else { res.add(p); }  
    } 

    if (!src.isComposition() && trg.isComposition()) 
    { CorrelationPattern p = new CorrelationPattern("Introduce aggregation", 
                                                    "Non-aggregation " + realsrc + "::" + src + 
                                                    " maps to aggregation " + 
                                                    realtrg + "::" + trg); 
      p.addSourceEntity(realsrc); 
      p.addTargetEntity(realtrg); 
      p.addSourceFeature(src); 
      p.addTargetFeature(trg); 
      if (res.contains(p)) { } 
      else { res.add(p); }  
    } 

    if (src.getNavigation().size() <= 1 && 
        trg.getNavigation().size() > 1) 
    { CorrelationPattern p = new CorrelationPattern("Introduce indirection", 
                                                    "Direct feature " + realsrc + "::" + src + 
                                                    " maps to composed feature " + 
                                                    realtrg + "::" + trg); 
      p.addSourceEntity(realsrc); 
      p.addTargetEntity(realtrg); 
      p.addSourceFeature(src); 
      p.addTargetFeature(trg); 
      if (res.contains(p)) { } 
      else { res.add(p); }  
    } 


    if (src.getNavigation().size() > 1 && 
        trg.getNavigation().size() <= 1) 
    { CorrelationPattern p = new CorrelationPattern("Remove indirection", 
                                                    "Composed feature " + realsrc + "::" + src + 
                                                    " maps to direct feature " + realtrg + "::" + trg); 
      p.addSourceEntity(realsrc); 
      p.addTargetEntity(realtrg); 
      p.addSourceFeature(src); 
      p.addTargetFeature(trg); 
      if (res.contains(p)) { } else { res.add(p); }  
    } 

    if (src.isComposed()) 
    { Attribute asrc = src.getFinalFeature(); 
      Attribute atrg = null; 
      if (trg.isComposed())
      { atrg = trg.getFinalFeature(); } 
      else 
      { atrg = trg; } 
      Entity own1 = asrc.getOwner(); 
      Entity own2 = atrg.getOwner(); 
      // System.out.println("OWNER of " + asrc + " is " + own1); 
      // System.out.println("OWNER of " + atrg + " is " + own2); 
      Entity srcimg = ModelMatching.lookupRealMatch(own1,ems); 
      // System.out.println("IMAGE of " + own1 + " is " + srcimg); 
      if (srcimg != null && srcimg != own2)
      { CorrelationPattern mq; 
        if (Entity.isAncestor(own2,srcimg))
        { mq = new CorrelationPattern("Pull up feature", 
                                      "Feature " + own1 + "::" + asrc + " is moved to superclass " + 
                                      own2 + "::" + atrg + " of " + srcimg); 
        } 
        else if (Entity.isAncestor(srcimg,own2))
        { mq = new CorrelationPattern("Push down feature", 
                                      "Feature " + own1 + "::" + asrc + " is moved to subclass " + 
                                      own2 + "::" + atrg + " of " + srcimg); 
        } 
        else 
        { mq = new CorrelationPattern("Feature moved", 
                                      "Feature " + own1 + "::" + asrc + " is moved to " + 
                                      own2 + "::" + atrg); 
        } 
        mq.addSourceEntity(own1); 
        mq.addTargetEntity(own2); 
        mq.addTargetEntity(srcimg); 
        mq.addSourceFeature(asrc); 
        mq.addTargetFeature(atrg); 
        if (res.contains(mq)) { } else { res.add(mq); }  
      } 
    } 
    else if (trg.isComposed()) 
    { Attribute atrg = trg.getFinalFeature(); 
      Attribute asrc = null; 
      if (src.isComposed())
      { asrc = src.getFinalFeature(); } 
      else 
      { asrc = src; } 
      Entity own1 = asrc.getOwner(); 
      Entity own2 = atrg.getOwner(); 
      // System.out.println("OWNER of " + asrc + " is " + own1); 
      // System.out.println("OWNER of " + atrg + " is " + own2); 
      Entity srcimg = ModelMatching.lookupRealMatch(own1,ems); 
      // System.out.println("IMAGE of " + own1 + " is " + srcimg); 
      if (srcimg != null && srcimg != own2)
      { CorrelationPattern mq; 
        if (Entity.isAncestor(own2,srcimg))
        { mq = new CorrelationPattern("Pull up feature", 
                                      "Feature " + own1 + "::" + asrc + " is moved to superclass " + 
                                      own2 + "::" + atrg + " of " + srcimg); 
        } 
        else if (Entity.isAncestor(srcimg,own2))
        { mq = new CorrelationPattern("Push down feature", 
                                      "Feature " + own1 + "::" + asrc + " is moved to subclass " + 
                                      own2 + "::" + atrg + " of " + srcimg); 
        }
        else 
        { mq = new CorrelationPattern("Feature moved", 
                                      "Feature " + own1 + "::" + asrc + " is moved to " + 
                                      own2 + "::" + atrg);
        }  
        mq.addSourceEntity(own1); 
        mq.addTargetEntity(own2); 
        mq.addTargetEntity(srcimg); 
        mq.addSourceFeature(asrc); 
        mq.addTargetFeature(atrg); 
        if (res.contains(mq)) { } else { res.add(mq); } 
      } 
    } 
    else // both direct, can still be moved
    { Entity own1 = src.getOwner(); 
      Entity own2 = trg.getOwner(); 
      // System.out.println("OWNER of " + asrc + " is " + own1); 
      // System.out.println("OWNER of " + atrg + " is " + own2); 
      Entity srcimg = ModelMatching.lookupRealMatch(own1,ems); 
      // System.out.println("IMAGE of " + own1 + " is " + srcimg); 
      if (srcimg != null && srcimg != own2)
      { CorrelationPattern mq; 
        if (Entity.isAncestor(own2,srcimg))
        { mq = new CorrelationPattern("Pull up feature", 
                                      "Feature " + own1 + "::" + src + " is moved to superclass " + 
                                      own2 + "::" + trg + " of " + srcimg); 
        } 
        else if (Entity.isAncestor(srcimg,own2))
        { mq = new CorrelationPattern("Push down feature", 
                                      "Feature " + own1 + "::" + src + " is moved to subclass " + 
                                      own2 + "::" + trg + " of " + srcimg); 
        }
        else 
        { mq = new CorrelationPattern("Feature moved", 
                                      "Feature " + own1 + "::" + src + " is moved to " + 
                                      own2 + "::" + trg);
        }  
        mq.addSourceEntity(own1); 
        mq.addTargetEntity(own2); 
        mq.addTargetEntity(srcimg); 
        mq.addSourceFeature(src); 
        mq.addTargetFeature(trg); 
        if (res.contains(mq)) { } else { res.add(mq); } 
      } 
    } 


    Type srct = src.getType(); 
    Type trgt = trg.getType(); 
    if (("" + srct).equals("" + trgt)) { } 
    else if (Type.isRecursiveSubtype(srct,trgt))
    { CorrelationPattern q = new CorrelationPattern("Widen type", 
                                   "Feature " + realsrc + "::" + src +
                                   " type widened from " + srct + " to " + trgt + " in " +  
                                   realtrg + "::" + trg); 
      q.addSourceEntity(realsrc); 
      q.addTargetEntity(realtrg); 
      q.addSourceFeature(src); 
      q.addTargetFeature(trg); 
      if (res.contains(q)) { } else { res.add(q); }  
    }  
    else if (Type.isRecursiveSubtype(trgt,srct))
    { CorrelationPattern q = new CorrelationPattern("Narrow type", 
                                   "Feature " + realsrc + "::" + src +
                                   " type narrowed from " + srct + " to " + trgt + " in " +  
                                   realtrg + "::" + trg); 
      q.addSourceEntity(realsrc); 
      q.addTargetEntity(realtrg); 
      q.addSourceFeature(src); 
      q.addTargetFeature(trg); 
      if (res.contains(q)) { } else { res.add(q); }  
    }  

    int mult1 = srct.typeMultiplicity(); 
    int mult2 = trgt.typeMultiplicity(); 
    if (mult1 == ModelElement.ONE && mult2 == ModelElement.MANY) 
    { CorrelationPattern q = new CorrelationPattern("Widen multiplicity", 
                                   "Feature " + realsrc + "::" + src +
                                   " widened from ONE to MANY multiplicity in " + 
                                   realtrg + "::" + trg); 
      q.addSourceEntity(realsrc); 
      q.addTargetEntity(realtrg); 
      q.addSourceFeature(src); 
      q.addTargetFeature(trg); 
      if (res.contains(q)) { } else { res.add(q); }  
    } 
    else if (mult1 == ModelElement.MANY && mult2 == ModelElement.ONE) 
    { CorrelationPattern q = new CorrelationPattern("Narrow multiplicity", 
                                   "Feature " + realsrc + "::" + src +
                                   " narrowed from MANY to ONE multiplicity in " + 
                                   realtrg + "::" + trg); 
      q.addSourceEntity(realsrc); 
      q.addTargetEntity(realtrg); 
      q.addSourceFeature(src); 
      q.addTargetFeature(trg); 
      if (res.contains(q)) { } else { res.add(q); }  
    } 

     
    return res; 
  } // also feature merging; stereotype changes


  public EntityMatching isObjectMatch(Vector ems) 
  { Vector spath = new Vector(); 
    spath.addAll(src.getNavigation()); 
    if (spath.size() == 0) 
    { spath.add(src); } 
    Vector tpath = new Vector(); 
    tpath.addAll(trg.getNavigation()); 
    if (tpath.size() == 0) 
    { tpath.add(trg); } 

    // Attribute s1 = (Attribute) spath.get(0); 
    // Attribute t1 = (Attribute) tpath.get(0); 

    

    Type tsrc = src.getElementType(); 
    Type ttrg = trg.getElementType(); 

    // System.out.println(">_>_ " + src + " (" + tsrc + ") " + trg + " (" + ttrg + 
    //                    ")"); 

    if (Type.isEntityType(tsrc) && Type.isEntityType(ttrg))
    { Entity e1 = tsrc.getEntity(); 
      Entity e2 = ttrg.getEntity();

      EntityMatching emx = ModelMatching.findEntityMatching(e1,ems); 
      if (emx != null && emx.realtrg != null) 
      { if (e2.getName().equals(emx.realtrg.getName()) || 
            Entity.isAncestor(e2,emx.realtrg))
        { return emx; } 
      } // to map from e1 to e2, emx.realsrc must be e1 or an ancestor of e1,
        // emx.realtrg must be e2 or a descendent of e2
    }  
    return null; 
  } 

  String whenClause(EntityMatching emx, String srcroot, String trgroot, Map whens) 
  { // isObjectMatching is true
    Vector spath = new Vector(); 
    spath.addAll(src.getNavigation()); 
    if (spath.size() == 0) 
    { spath.add(src); } 
    Vector tpath = new Vector(); 
    tpath.addAll(trg.getNavigation()); 
    if (tpath.size() == 0) 
    { tpath.add(trg); } 

    // Attribute s1 = (Attribute) spath.get(0); 
    // Attribute t1 = (Attribute) tpath.get(0); 
    // Type tsrc = src.getElementType(); 
    // Type ttrg = trg.getElementType(); 
    String nsrc = emx.src.getName(); 
    String ntrg = emx.trg.getName(); 
    String srcentx = nsrc.substring(0,nsrc.length()-1); 
    if (emx.realsrc != null) 
    { srcentx = emx.realsrc.getName(); } 
    String trgentx = ntrg.substring(0,ntrg.length()-1); 
    if (emx.realtrg != null) 
    { trgentx = emx.realtrg.getName(); } 

    String d1 = dataname(srcroot,spath); 
    String d2 = dataname(trgroot,tpath); 

    BasicExpression be1 = new BasicExpression(d1); 
    be1.setType(src.getType()); 
    be1.setElementType(new Type(emx.realsrc)); 
    be1.variable = src; 

    BasicExpression be2 = new BasicExpression(d2); 
    be2.setType(trg.getType()); 
    be2.setElementType(new Type(emx.realtrg)); 
    be2.variable = trg; 

    whens.set(be2,be1); 
    return srcentx + "2" + trgentx + "(" + d1 + "," + d2 + ")"; 
  } 

  String whenClause(String trgroot, Vector ems, Map whens) 
  { // For expr --> trg, need E2F(elementVariable,d2) 
    if (elementVariable == null) 
    { return ""; } 

    Vector tpath = new Vector(); 
    tpath.addAll(trg.getNavigation()); 
    if (tpath.size() == 0) 
    { tpath.add(trg); } 

    Type vtype = elementVariable.getElementType(); 
    if (vtype != null && vtype.isEntity())
    { Entity vent = vtype.getEntity(); 
      System.out.println(">>> Entity of " + elementVariable + " is " + vent); 

      Type ttype = trg.getElementType(); 
      if (ttype != null && ttype.isEntity())
      { Entity tent = ttype.getEntity(); 
        EntityMatching emx = ModelMatching.findEntityMatching(vent,ems); 
        System.out.println(">>> Entity match of " + elementVariable + " is " + emx); 
        if (emx != null) 
        { String nsrc = emx.src.getName(); 
          String ntrg = emx.trg.getName(); 
          String srcentx = nsrc.substring(0,nsrc.length()-1); 
          if (emx.realsrc != null) 
          { srcentx = emx.realsrc.getName(); } 
          String trgentx = ntrg.substring(0,ntrg.length()-1); 
          if (emx.realtrg != null) 
          { trgentx = emx.realtrg.getName(); } 

          String d2 = dataname(trgroot,tpath); 

          // BasicExpression be2 = new BasicExpression(d2); 
          // be2.setType(trg.getType()); 
          // be2.setElementType(new Type(emx.realtrg)); 
          // be2.variable = trg; 

          // whens.set(be2,be1); 
          return srcentx + "2" + trgentx + "(" + elementVariable + "," + d2 + ")"; 
        } 
      } 
    } 
    return ""; 
  } 

  String sourcedataname(String prefix)
  { Vector path = new Vector(); 
    path.addAll(src.getNavigation()); 
    if (path.size() == 0) 
    { path.add(src); }
    return dataname(prefix,path); 
  } 

  static String dataname(String prefix, Attribute att)
  { Vector path = new Vector(); 
    path.addAll(att.getNavigation()); 
    if (path.size() == 0) 
    { path.add(att); }
    return dataname(prefix,path); 
  } 

  static String dataname(String prefix, Vector path)
  { // nme = prefix + "_" + nme$value for non-objects, non-collections
    // nme = prefix + "_" + nme$x for objects, collections

    if (path.size() == 1) 
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType();
      String pname = p.getName(); 
      String fullname = prefix + "_" + pname; 
 
      if (t != null && Type.isCollectionOrEntity(t))
      { Type telem = t.getElementType(); 
        if ((telem != null && telem.isEntity()) || t.isEntity())
        { return fullname + "$x"; }
      }  
      return fullname + "$value";
    }  
    else if (path.size() > 1) 
    { Attribute p = (Attribute) path.get(0); 
      String pname = p.getName(); 
      String fullname = prefix + "_" + pname; 
      Vector pathtail = new Vector(); 
      pathtail.addAll(path); 
      pathtail.remove(0); 
 
      return dataname(fullname, pathtail); 
    }      

    // String pname = src.getName(); 
    // String fullname = prefix + "_" + pname; 
    // return fullname + "$value";
    return prefix;  
  } 

  String sourceequation(String prefix, ObjectTemplateExp sexp)
  { // nme = nme$value for non-objects, non-collections
    Vector path = new Vector(); 
    path.addAll(src.getNavigation()); 
    if (path.size() == 0) 
    { path.add(src); }
    return sourceequation(prefix,path,sexp); 
  } 

  String sourceequation(String prefix, Vector path, ObjectTemplateExp objTE)
  { // nme = nme$value for non-objects, non-collections

    if (path.size() == 1) 
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType();
      String pname = p.getName(); 
      String fullname = prefix + "_" + pname; 
 
      if (t != null && Type.isCollectionOrEntity(t))
      { Type telem = t.getElementType(); 
        if ((telem != null && telem.isEntity()) || t.isEntity())
        { Entity tent; 
          if (telem == null) 
          { tent = t.getEntity(); } 
          else 
          { tent = telem.getEntity(); } 

          String obj = fullname + "$x";
          Attribute objroot = new Attribute(obj,new Type(tent),ModelElement.INTERNAL); 
          objroot.setElementType(new Type(tent)); 

          // String defn = pname + " = " + obj + " : " + tent; 
 
          // Vector exps = (Vector) objectTemplates.get(defn);
          // if (exps == null) 
          // { exps = new Vector(); }
          
          // objectTemplates.set(defn,exps);
          // return "";
          objTE.addPTI(p,new ObjectTemplateExp(objroot,tent));     
          return pname + " = " + obj + " : " + tent + " { }";
        }  
      } 
      BasicExpression valueExp = new BasicExpression(fullname + "$value"); 
      valueExp.setType(p.getType()); 
      valueExp.setElementType(p.getElementType()); 
      valueExp.variable = src; // to record the actual thing. 

      objTE.addPTI(p,valueExp);     
      return pname + " = " + fullname + "$value ";
    }  
    else if (path.size() > 1) 
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType();
      String pname = p.getName(); 
      String fullname = prefix + "_" + pname; 
 
      // (Type.isCollectionOrEntity(t))
      Entity tent = null; 
      String obj = fullname + "$x"; 

      Type telem = t.getElementType(); 
      if (telem != null && telem.isEntity())
      { tent = telem.getEntity(); } 
      else if (t.isEntity())
      { tent = t.getEntity(); } 

      if (tent != null)  
      { Vector pathtail = new Vector(); 
        pathtail.addAll(path); 
        pathtail.remove(0); 

        String body = "";         
        Object subtemplate = objTE.getPTI(p); 
        if (subtemplate instanceof ObjectTemplateExp)
        { ObjectTemplateExp subTE = (ObjectTemplateExp) subtemplate; 
          body = sourceequation(prefix + "_" + pname, pathtail, subTE); 

        // String defn = pname + " = " + obj + " : " + tent; 
        // Vector exps = (Vector) objectTemplates.get(defn);
        // if (exps == null) 
        // { exps = new Vector(); }
        // exps.add(body); 
        // objectTemplates.set(defn,exps);   

          return pname + " = " + obj + " : " + tent + " { " + body + " }";
        } 
        else 
        { Attribute objroot = new Attribute(obj,new Type(tent),ModelElement.INTERNAL); 
          objroot.setElementType(new Type(tent)); 

          ObjectTemplateExp newSubTE = new ObjectTemplateExp(objroot,tent); 
          objTE.addPTI(p,newSubTE);     
          body = sourceequation(prefix + "_" + pname, pathtail, newSubTE); 
          return pname + " = " + obj + " : " + tent + " { " + body + " }";
        }   
      }  

      BasicExpression valueExp = new BasicExpression(fullname + "$value"); 
      valueExp.setType(p.getType()); 
      valueExp.setElementType(p.getElementType()); 
      valueExp.variable = src; 

      objTE.addPTI(p,valueExp);     
      return pname + " = " + fullname + "$value";
    }      

    // String pname = src.getName(); 
    // String fullname = prefix + "_" + pname; 
    // return pname + " = " + fullname + "$value"; 
    return ""; 
  } 

  String targetequation(String prefix,String srcdata,ObjectTemplateExp tExps)
  { // nme = nme$value for non-objects, non-collections


    Vector path = new Vector(); 
    path.addAll(trg.getNavigation()); 
    if (path.size() == 0) 
    { path.add(trg); }
    return targetequation(prefix,path,srcdata,tExps); 
  } 

  String targetequationUMLRSDS(String tvar, Vector bound)
  { Vector path = new Vector(); 
    path.addAll(trg.getNavigation()); 
    if (path.size() == 0)
    { path.add(trg); } 
    return targetequationUMLRSDS(tvar,path,bound); 
  } 

  String targetequation(String prefix, Vector path, String srcdata, ObjectTemplateExp tExp)
  { // nme = srcdata at the end of the path

    if (path.size() == 1) 
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType();
      String pname = p.getName(); 
      String fullname = prefix + "_" + pname; 
 
      if (Type.isCollectionOrEntity(t))
      { Entity tent = null; 
        Type telem = t.getElementType(); 
        if (telem != null && telem.isEntity())
        { tent = telem.getEntity(); } 
        else if (t.isEntity())
        { tent = t.getEntity(); } 

        if (tent != null) 
        { // String defn = pname + " = " + fullname + "$x : " + tent;  
          // Vector exps = (Vector) tExps.get(defn);
          // if (exps == null) 
          // { exps = new Vector(); }
          
          // tExps.set(defn,exps);
          // return "";    

          String obj = fullname + "$x";
          Attribute objroot = new Attribute(obj,new Type(tent),ModelElement.INTERNAL); 
          objroot.setElementType(new Type(tent)); 

          tExp.addPTI(p,new ObjectTemplateExp(objroot,tent));     
          return pname + " = " + fullname + "$x : " + tent + " { }"; 
        } 
      }
      BasicExpression valueExp = new BasicExpression(srcdata); 
      valueExp.setType(p.getType()); 
      valueExp.setElementType(p.getElementType()); 
      valueExp.variable = src; 

      Expression valueExpC = valueExp; 
      if ("String".equals(p.getType() + "") && src.getType().isEnumeration())
      { valueExpC = AttributeMatching.dataConversion(valueExp, src.getType(), p.getType()); } 
      else if ("String".equals(src.getType() + "") && p.getType().isEnumeration())
      { valueExpC = AttributeMatching.dataConversion(valueExp, src.getType(), p.getType()); } 

      tExp.addPTI(p,valueExpC);     
      return pname + " = " + valueExpC; 
    }  
    else if (path.size() > 1) 
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType();
      String pname = p.getName(); 
      String fullname = prefix + "_" + pname; 
 
      // (Type.isCollectionOrEntity(t))
      Type telem = t.getElementType(); 
      Entity tent = telem.getEntity(); 
      if (tent == null) 
      { tent = t.getEntity(); } 

      Vector pathtail = new Vector(); 
      pathtail.addAll(path); 
      pathtail.remove(0); 
      String body = ""; 
      // String defn = pname + " = " + fullname + "$x : " + tent; 

      Object subtemplate = tExp.getPTI(p); 
      if (subtemplate instanceof ObjectTemplateExp)
      { ObjectTemplateExp subTE = (ObjectTemplateExp) subtemplate; 
        body = targetequation(prefix + "_" + pname, pathtail, srcdata, subTE); 
      } 
      else 
      { String obj = fullname + "$x"; 
        Attribute objroot = new Attribute(obj,new Type(tent),ModelElement.INTERNAL); 
        objroot.setElementType(new Type(tent)); 

        ObjectTemplateExp newSubTE = new ObjectTemplateExp(objroot,tent); 
        tExp.addPTI(p,newSubTE);     
        body = targetequation(prefix + "_" + pname, pathtail, srcdata, newSubTE); 
        return pname + " = " + obj + " : " + tent + " { " + body + " }";
      }   
      // return pname + " = " + fullname + "$x : " + tent + " { " + body + " }"; 
    }      

    if (path.size() > 0) 
    { Attribute p = (Attribute) path.get(0); 
      String pname = p.getName(); 
      BasicExpression valueExp = new BasicExpression(srcdata); 
      valueExp.setType(p.getType()); 
      valueExp.setElementType(p.getElementType()); 
      valueExp.variable = src; 

      Expression valueExpC = valueExp; 
      if ("String".equals(p.getType() + "") && src.getType().isEnumeration())
      { valueExpC = AttributeMatching.dataConversion(valueExp, src.getType(), p.getType()); } 
      else if ("String".equals(src.getType() + "") && p.getType().isEnumeration())
      { valueExpC = AttributeMatching.dataConversion(valueExp, src.getType(), p.getType()); } 

      tExp.addPTI(p,valueExpC);     
      return pname + " = " + valueExpC;
    } 
    return "";  
  } 

  String targetequationUMLRSDS(String tvar, Vector path, Vector bound) 
  { Type ttarg = trg.getType(); 
    Type tsrc = src.getType(); 

    String res = ""; 
    if (ttarg == null) 
    { System.err.println("ERROR: null type for " + trg); 
      return res; 
    } 
    if (tsrc == null) 
    { System.err.println("ERROR: null type for " + src); 
      return res; 
    } 

    int multsrc = tsrc.typeMultiplicity(); 
    int multtrg = ttarg.typeMultiplicity(); 
 
    if (path.size() == 1) 
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType(); 

      if (t != null && Type.isEntityCollection(t))
      { // convert the target objects
       
        Type telem = t.getElementType();

        Entity tent = telem.getEntity(); 
        String tename = tent.getName(); 
        Type selem = tsrc.getElementType(); // must also be an entity

        Entity sentity = selem.getEntity(); 
        String sename = sentity.getName(); 
        String sId = sename.toLowerCase() + "Id"; 

        if (multsrc != ModelElement.ONE && multtrg == ModelElement.ONE)
        { res = tvar + "." + trg + " = " + tename + "[" + src + "." + sId + "]->any()"; } 
        else if (multsrc == ModelElement.ONE && multtrg != ModelElement.ONE)
        { res = tename + "[" + src + "." + sId + "] : " + tvar + "." + trg; }  
        else if ("Sequence".equals(t.getName()) && "Set".equals(src.getType().getName()))
        { res = tvar + "." + trg + " = " + tename + "[" + src + "." + sId + "]->asSequence()"; } 
        else if ("Set".equals(t.getName()) && "Sequence".equals(src.getType().getName()))
        { res = tvar + "." + trg + " = " + tename + "[" + src + "." + sId + "]->asSet()"; } 
        else 
        { res = tvar + "." + trg + " = " + tename + "[" + src + "." + sId + "]"; } 
      } 
      else if (t != null && t.isEntity())
      { // convert the target objects

        Entity tent = t.getEntity(); 
        String tename = tent.getName(); 
      
        Entity sentity = tsrc.getEntity(); 
        if (sentity == null) 
        { Type tsrcelem = src.getElementType(); 
          sentity = tsrcelem.getEntity(); 
        } 

        if (sentity == null) 
        { System.err.println("ERROR: not entity type/element type: " + tsrc); 
          return ""; 
        } 

        String sename = sentity.getName(); 
        String sId = sename.toLowerCase() + "Id"; 
        if (multsrc != ModelElement.ONE && multtrg == ModelElement.ONE)
        { res = tvar + "." + trg + " = " + tename + "[" + src + "." + sId + "]->any()"; } 
        else if (multsrc == ModelElement.ONE && multtrg != ModelElement.ONE)
        { res = tename + "[" + src + "." + sId + "] : " + tvar + "." + trg; }  
        else 
        { res = tvar + "." + trg + " = " + tename + "[" + src + "." + sId + "]"; }         
      } 
      else if (multsrc != ModelElement.ONE && multtrg == ModelElement.ONE)
      { res = tvar + "." + trg + " = " + src + "->any()"; } 
      else if (multsrc == ModelElement.ONE && multtrg != ModelElement.ONE)
      { res = src + " : " + tvar + "." + trg; }  
      else 
      { BasicExpression sexp = new BasicExpression(src); 
        Expression texp = AttributeMatching.dataConversion(sexp,tsrc,ttarg); 
        res = tvar + "." + trg + " = " + texp;
      } 
    } 
    return res; 
  } 
 
  String composedTargetEquation(String tvar, Vector created)
  { Vector path = new Vector(); 
    path.addAll(trg.getNavigation()); 
    if (path.size() == 0)
    { path.add(trg); } 

    boolean alreadycreated = false; 
    boolean multiple = false; 

    Vector preattp = VectorUtil.largestInitialSegment(path,created);
    Attribute preatt = null; 
    String prex = "";
    String tvar1 = tvar;  
    Vector pathrem = new Vector(); 
 
    if (preattp != null && preattp.size() > 0) 
    { preatt = new Attribute(preattp); 
      prex = preatt.getName(); // ModelElement.composeNames(preatt); 
      tvar1 = tvar + "." + prex; 
      pathrem.addAll(path); 
      pathrem.removeAll(preattp); 
      // System.out.println(">>> Already created " + preatt + " " + tvar1 + " " + pathrem); 
      alreadycreated = true; 
      multiple = preatt.isManyValued(); 
    } 

    Type t = trg.getType(); 
    Entity e = null; 
    if (Type.isEntityType(t))
    { e = t.getEntity(); } 
    else if (Type.isEntityCollection(t))
    { Type telem = t.getElementType(); 
      e = telem.getEntity(); 
    } 

    if (e != null)  
    { String ename = e.getName();  
      String sname = "";  
      String bname = ""; 
      // String shortename = ename.substring(0,ename.length()-1);  // target feature element type 
      Type stype = src.getElementType(); 
      if (Type.isEntityType(stype))
      { Entity b = stype.getEntity(); 
        bname = b.getName(); 
        // sname = bname.substring(0,bname.length()-1);  // source feature element type
      }  
      String bId = bname.toLowerCase() + "Id"; 
      String evar = ename.toLowerCase() + "$x"; 
      String evardec = ""; 
      if (Type.isCollectionType(src.getType()))
      { evardec = " & " + evar + " : " + ename + "[" + src + "." + bId + "] => \n"; } 
      else 
      { evardec = " & " + evar + " = " + ename + "[" + src + "." + bId + "] => \n"; }
      if (alreadycreated)
      { if (multiple) 
        { return evardec + "      " + tvar1 + "->forAll( _y | " + 
                                  composedTargetEquation("_y",evar,pathrem) + " )"; 
        } 
        else 
        { String cteq = composedTargetEquation(tvar1,evar,pathrem);
          if (cteq.length() > 0) 
          { return evardec + "      " + cteq; } 
          return "";                
        } 
      } 
      String cteq1 = composedTargetEquation(tvar,evar,path);
      if (cteq1.length() > 0)
      { return evardec + "      " + cteq1; } 
      return "";  
    } 
    else // target is of a value type 
    if (Type.isCollectionType(src.getType()))
    { String evar = Identifier.nextIdentifier("var$"); 
      if (alreadycreated)
      { if (multiple) 
        { return " & " + evar + " : " + src + " =>\n" + 
             "      " + tvar1 + "->forAll( _y | " + 
                        composedTargetEquation("_y",evar,pathrem) + " )"; 
        } 
        else 
        { String cteq = composedTargetEquation(tvar1,evar,pathrem);
          if (cteq.length() > 0) 
          { return " & " + evar + " : " + src + " =>\n" + 
                   "      " + cteq; 
          } 
          return "";
        } 
      } 
      String cteq1 = composedTargetEquation(tvar,evar,path);
      if (cteq1.length() > 0)
      { return " & " + evar + " : " + src + " =>\n" + 
                "      " + cteq1; 
      } 
      return "";  
    } 
    else 
    { String evar = src + ""; 
      if (alreadycreated)
      { if (multiple) 
        { return " =>\n" + 
           "       " + tvar1 + "->forAll( _y | " + 
                       composedTargetEquation("_y",evar,pathrem) + " )";
        }
        else 
        { String cteq = composedTargetEquation(tvar1,evar,pathrem);
          if (cteq.length() > 0)
          { return " =>\n" + 
               "       " + cteq;
          } 
          return ""; 
        }
      }  
      String cteq1 = composedTargetEquation(tvar,evar,path);
      if (cteq1.length() > 0)
      { return " =>\n" + "       " + cteq1; } 
      return "";  
    } 
    // return tvar + "." + trg + " = " + src;  
  } 

  public String composedTargetEquation(String svar, String evar, Vector path) 
  { if (path.size() == 1) 
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType(); 
      if (Type.isCollectionType(t))
      { return evar + " : " + svar + "." + p.getName(); } 
      else if (t.isEnumeration() && src.getType().getName().equals("String"))
      { BasicExpression sexp = new BasicExpression(evar); 
        Type tsrc = src.getType(); 
        Expression texp = AttributeMatching.dataConversion(sexp,tsrc,t); 
      } 
      else if (src.getType().isEnumeration() && t.getName().equals("String"))
      { BasicExpression sexp = new BasicExpression(evar); 
        Type tsrc = src.getType(); 
        Expression texp = AttributeMatching.dataConversion(sexp,tsrc,t); 
      } 
      else 
      { return svar + "." + p.getName() + " = " + evar; } 
    } 
    else if (path.size() > 0)
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType(); 
      Entity e = null; 
      if (Type.isEntityType(t))
      { e = t.getEntity(); } 
      else if (Type.isEntityCollection(t))
      { Type elemT = t.getElementType(); 
        e = elemT.getEntity(); 
      } 
      String ename = e.getName(); 
      String ex = ename.toLowerCase() + "$x"; 
      Vector ptail = new Vector(); 
      ptail.addAll(path); 
      ptail.remove(0); 
      String code = composedTargetEquation(ex,evar,ptail); 
      if (Type.isCollectionType(t))
      { code = ex + " : " + svar + "." + p.getName() + " & " + code; } 
      else 
      { code = svar + "." + p.getName() + " = " + ex + " & " + code; } 
      return ename + "->exists( " + ex + " | " + code + " ) "; 
    } 
    return ""; 
  } // push down the evar : ename[src.bId] declaration as locally as possible. 

  String targetequationQVTO(Expression tvar, Expression srcdata, Vector ems, Vector created)
  { Vector path = new Vector(); 
    path.addAll(trg.getNavigation()); 
    if (path.size() == 0)
    { path.add(trg); } 

    Vector preattp = VectorUtil.largestInitialSegment(path,created); 
    if (preattp != null && preattp.size() > 0) 
    { Attribute preatt = new Attribute(preattp); 
      String prex = preatt.getName(); // ModelElement.composeNames(preatt); 
      String tvar1 = tvar + "." + prex; 
      Vector pathrem = new Vector(); 
      pathrem.addAll(path); 
      pathrem.removeAll(preattp); 
      // System.out.println(">>> Already created " + preatt + " " + tvar1 + " " + pathrem); 
      if (preatt.isManyValued())
      { BasicExpression y = new BasicExpression("_y"); 
        y.setType(preatt.getElementType()); 
        String body = targetequationQVTO(y,srcdata,pathrem,ems,created); 
        return tvar1 + "->forEach(_y) { " + body + " };"; 
      } 
      BasicExpression p = new BasicExpression(preatt); 
      p.setObjectRef(tvar); 
      return targetequationQVTO(p,srcdata,pathrem,ems,created); 
      // tvar = result.preatt if ONE mult, path is remainder of path
    } 

    return targetequationQVTO(tvar,srcdata,path,ems,created); 
  } 

  String targetequationQVTO(Expression tvar, Expression svar, Vector path, Vector ems, Vector created) 
  { Type ttarg = trg.getType(); 
    Type tsrc = src.getType(); 
    int smult = tsrc.typeMultiplicity(); 
    int tmult = ttarg.typeMultiplicity(); 

    String res = ""; 
 
    if (path.size() == 1) 
    { Attribute p = (Attribute) path.get(0); 
      Type t = p.getType(); 
      if (t != null) 
      { tmult = t.typeMultiplicity(); } 

      if (t != null && Type.isEntityCollection(t))
      { // convert the target objects using resolve
        // unless the target entity element type is shared between source & target models. 
        // Also take account of the multiplicities. 
        Type telem = t.getElementType();

        Entity tent = telem.getEntity();
        if (tent != null && !tent.isSourceEntity() && !tent.isTargetEntity())
        { res = tvar + "." + p + " := " + svar + "." + src; } 
        else if (tent != null)  
        { String tename = tent.getName(); 
          Type selem = tsrc.getElementType(); // must also be an entity
          Entity s = tsrc.getEntity(); 
          if (selem != null && selem.isEntity())
          { s = selem.getEntity(); } 

          EntityMatching em = ModelMatching.findEntityMatching(s,ems);
          // System.out.println("Entity matching for " + s + " is " + em); 

          if (em != null && smult == ModelElement.ONE) 
          { res = tvar + "." + p + " := " + svar + "." + src + 
                 ".resolveoneIn(" + em.realsrc + "::" + 
                 em.realsrc + "2" + em.realtrg + "," + tename + ")";
          } 
          else if (em != null && smult != ModelElement.ONE)
          { res = tvar + "." + p + " := " + svar + "." + src + 
                 ".resolveIn(" + em.realsrc + "::" + 
                 em.realsrc + "2" + em.realtrg + "," + tename + ")";
          } 
          else 
          { res = tvar + "." + p + " := " + svar + "." + src; }  
        } 
      } 
      else if (t != null && t.isEntity())
      { // convert the target objects using resolveone

        Entity tent = t.getEntity(); 
      
        if (tent != null && !tent.isSourceEntity() && !tent.isTargetEntity())
        { res = tvar + "." + p + " := " + svar + "." + src; } 
        else if (tent != null)  
        { String tename = tent.getName(); 
          Type selem = tsrc.getElementType(); // must also be an entity
          Entity s = tsrc.getEntity(); 
          if (selem != null && selem.isEntity())
          { s = selem.getEntity(); } 
          
          EntityMatching em = ModelMatching.findEntityMatching(s,ems); 
          // System.out.println("Entity matching for " + s + " is " + em); 

          if (em != null && smult == ModelElement.ONE) 
          { res = tvar + "." + p + " := " + svar + "." + src + 
                 ".resolveoneIn(" + em.realsrc + "::" + 
                 em.realsrc + "2" + em.realtrg + "," + tename + ")";
          } 
          else if (em != null && smult != ModelElement.ONE)
          { res = tvar + "." + p + " := " + svar + "." + src + 
                 ".resolveIn(" + em.realsrc + "::" + 
                 em.realsrc + "2" + em.realtrg + "," + tename + ")";
          } 
          else 
          { res = tvar + "." + p + " := " + svar + "." + src; }  
        } 
      } 
      else // it is a value or collection of values. 
      { BasicExpression srcexp = new BasicExpression(svar + "." + src);
        Expression sconv = AttributeMatching.dataConversion(srcexp, src.getType(), t);  
        res = tvar + "." + p + " := " + sconv + ";"; 
        // created.add(trg.getNavigation()); 
        return res; 
      }

      if (tmult == ModelElement.ONE && smult != ModelElement.ONE) 
      { res = res + "->any();"; } 
      else if (tmult != ModelElement.ONE && smult == ModelElement.ONE)
      { res = res + "->as" + t.getName() + "();"; } 
      else if (tmult != ModelElement.ONE && smult != ModelElement.ONE && 
               "Sequence".equals(t.getName()) && "Set".equals(src.getType().getName()))
      { res = res + "->asSequence();"; } 
      else if (tmult != ModelElement.ONE && smult != ModelElement.ONE && 
               "Set".equals(t.getName()) && "Sequence".equals(src.getType().getName()))
      { res = res + "->asSet();"; } 
      else 
      { res = res + ";"; } 

      // { res = AttributeMatching.dataConversion(new BasicExpression(res), tsrc, ttarg) + ";"; }  
      // other data conversions. 
      // created.add(trg.getNavigation()); 
    } 
    else if (path.size() > 1) 
    { Attribute p = (Attribute) path.get(0); 
      Vector pathtail = new Vector(); 
      pathtail.addAll(path); 
      pathtail.remove(0); 
      BasicExpression srcexp = new BasicExpression(src); 
      srcexp.setObjectRef(svar); // has same multiplicity and types as src
      BasicExpression trgexp = new BasicExpression(p); 
      trgexp.setObjectRef(tvar); 
      res = qvtoObjectFor(p, pathtail, srcexp, trgexp, ems, created); 
      // created.add(trg.getNavigation()); 
    }  
      // tvar + "." + trg + " := " + svar + "." + src + ";"; } 
    return res; 
  } 

  private String qvtoObjectFor(Attribute previous, Vector path, 
                               Expression svar, Expression tvar, 
                               Vector ems, Vector created)
  { String result = ""; 
    Type tsrc = src.getType(); 
    Type ttarg = trg.getType(); 

    boolean smult = svar.isMultiple(); 
    boolean tmult = tvar.isMultiple(); 
    String operator = ":="; 
    if (tmult) 
    { operator = "+="; } 

    if (path.size() == 1) 
    { Attribute p = (Attribute) path.get(0); 
      Entity t = p.getOwner(); 
      String pname = p.getName(); 
      Type ptype = p.getType(); 

      String pop = ":="; 
      Entity pent = null; 

      if (p.isManyValued())
      { pop = "+=";
        Type pelemt = p.getElementType(); 
        if (pelemt != null && pelemt.isEntity())
        { pent = pelemt.getEntity(); }
      } 
      else 
      { if (ptype.isEntity())
        { pent = ptype.getEntity(); } 
      } 

      Type selem = svar.getElementType(); // must also be an entity if ptype is
      if (selem == null) 
      { selem = svar.getType(); } 
        
      if (previous != null && previous.isManyValued() && src.isManyValued())
      { String xresolve = "_x.resolveone()"; 
        if (selem.isEntity())
        { EntityMatching emx = ModelMatching.findEntityMatching(selem.getEntity(), ems); 
          String desttype = ""; 
          if (emx != null) 
          { if (pent != null) 
            { desttype = pent.getName(); }  
            else 
            { desttype = emx.realtrg.getName(); } 
            xresolve = "_x.resolveoneIn(" + 
                       emx.realsrc + "::" + emx.realsrc + "2" + emx.realtrg + 
                       ", " + desttype + ")";
          }  
        } 
 
        if (ptype.getName().equals("Sequence"))
        { result = 
            tvar + " := " + svar + "->xcollect( _x | object " + 
                                        t + " { " + pname + " := Sequence{ " + xresolve + " }; } );"; 
        } 
        else if (ptype.getName().equals("Set"))
        { result = tvar + " := " + svar + "->xcollect( _x | object " + t + 
                                          " { " + pname + " := Set{ " + xresolve + " }; } );"; 
        } 
        else 
        { result = tvar + " := " + svar + "->xcollect( _x | object " + 
                                            t + " { " + pname + " := " + xresolve + "; } );"; 
        }
      } 
      else // dataconversions as for targetequationQVTO
      { 
        if (pent != null && selem != null && selem.isEntity()) 
        { Entity s = selem.getEntity(); 
          EntityMatching em = ModelMatching.findEntityMatching(s,ems); 
          // System.out.println("Entity matching for " + s + " is " + em); 

          String res = pname + " " + pop + " " + svar + "." + 
                 "resolveoneIn(" + em.realsrc + "::" + em.realsrc + "2" + em.realtrg + ", " +
                 pent.getName() + ");"; 
          result = tvar + " " + operator + " object " + t + " { " + res + " };"; 
        } 
        else 
        { result = tvar + " " + operator + " object " + t + 
                        " { " + pname + " " + pop + " " + svar + "; };"; 
        } 
      } 
    } 
    else if (path.size() > 1) 
    { Attribute p = (Attribute) path.get(0); 
      Entity t = p.getOwner();
      String pname = p.getName(); 
      Vector pathtail = new Vector(); 
      pathtail.addAll(path); 
      pathtail.remove(0); 

      BasicExpression xexp = new BasicExpression("_x"); 
      xexp.setType(svar.getElementType());  
      BasicExpression pexp = new BasicExpression(p); 

      if (previous != null && previous.isManyValued() && src.isManyValued())
      { String rem1 = qvtoObjectFor(null,pathtail,xexp,pexp,ems,created); 
        result = tvar + " := " + svar + "->xcollect( _x | object " + t + " { " + rem1 + " });"; 
      } 
      else   
      { String rem = qvtoObjectFor(p,pathtail,svar,pexp,ems,created); 
        result = tvar + " " + operator + " object " + t + " { " + rem + " };"; 
      } 
    } 
    return result; 
  } 

} 
